import importlib
import inspect
from os import path as os_path
from os import listdir
from sys import path as sys_path

__all__ = (
    "ModuleLoader"
)

_THIS_FILE = os_path.basename(__file__)
_DEFAULT_EXCLUDES = (
    '__init__.py',
    _THIS_FILE,
)
_DECORATOR_ATTR = "load_flg"


def _detect_call_path():
    for path in inspect.stack():
        path_name = path.filename
        filename = os_path.basename(path_name)
        if _THIS_FILE == filename:
            continue
        return path_name


def _init_base_url(base_path=None):
    if base_path is None:
        return _init_base_url(os_path.dirname(_detect_call_path()))
    if base_path == '/':
        return base_path
    if base_path == '':
        return '/'
    if base_path.endswith('/'):
        return base_path[:-1]
    return base_path


class ModuleLoader:
    def __init__(self, base_path=None):
        self.__base_path = _init_base_url(base_path)
        self.__context = None

    @property
    def base_path(self):
        return self.__base_path

    def load_class(self, file_name):
        self.__context = self.Context(self.Context.Type.clazz)
        return self.__load_resource(file_name)

    def load_function(self, file_name):
        self.__context = self.Context(self.Context.Type.func)
        return self.__load_resource(file_name)

    def load_classes(self, pkg_name, excludes=None):
        self.__context = self.Context(self.Context.Type.clazz)
        return self.__load_resources(pkg_name, excludes=excludes)

    def load_functions(self, pkg_name, excludes=None):
        self.__context = self.Context(self.Context.Type.func)
        return self.__load_resources(pkg_name, excludes=excludes, type='function')

    def __path_fix(self, name):
        if not name or name == '.' or name == '/' or name == './':
            return self.__base_path
        if name.startswith('/'):
            result_path = self.__base_path + name
            # example: /foo/bar/
            if name.endswith('/'):
                return result_path
            # example: /foo/bar
            return result_path + '/'
        if name.startswith('.'):
            if name[1] != '.':
                if name[1] == '/':
                    result_path = self.__base_path + name[1:]
                    # example: ./foo/
                    if name.endswith('/'):
                        return result_path
                    # example: ./foo
                    return result_path + '/'
                # example: .foo.bar
                return self.__base_path + '/'.join(name.split('.')) + '/'
            level = 0
            path = None
            for i in range(len(name)):
                if i == 0:
                    continue
                if name[i] != '.':
                    break
                path = name[i + 1:]
                level += 1
            base_path_arr = self.__base_path.split('/')
            result_base_path = "/".join(base_path_arr[0:len(base_path_arr) - level])
            if path.startswith('/'):
                if path.endswith('/'):
                    # example: ../foo/
                    return result_base_path + path
                # example: ../foo
                return result_base_path + path + '/'
            # example: ..foo.bar
            path = '/'.join(path.split('.'))
            return result_base_path + '/' + path + '/'
        # example: foo.bar
        path = '/'.join(name.split('.'))
        return self.__base_path + '/' + path + '/'

    def __load_resource(self, file_name):
        target_file = file_name.replace('.py', '') if file_name.endswith('.py') else file_name
        fix_path_arr = self.__path_fix(target_file).split('/')
        target_file = fix_path_arr[-2]
        target_path = '/'.join(fix_path_arr[:-2])
        if target_path not in sys_path:
            sys_path.append(target_path)
        module = importlib.import_module(target_file)
        comparison = self.__context.draw_comparison(target_file)
        for mod_name, resource in inspect.getmembers(module, self.__context.predicate):
            if hasattr(resource, _DECORATOR_ATTR) and resource.load_flg:
                return resource
            if comparison != mod_name.lower():
                continue
            del self.__context
            return resource

    def __load_resources(self, pkg_name, excludes=None, type='class'):
        target_dir = self.__path_fix(pkg_name)
        if not os_path.isdir(target_dir):
            raise NotADirectoryError('Not Found The Directory : {}'.format(target_dir))
        if target_dir not in sys_path:
            sys_path.append(target_dir)
        files = [os_path.splitext(file)[0] for file in listdir(target_dir) if file.endswith('.py')]
        exclude_files = list(_DEFAULT_EXCLUDES)
        exclude_files.append(os_path.basename(_detect_call_path()))
        if excludes:
            if not iter(excludes):
                raise TypeError('excludes variable must be iterable.')
            for exclude in excludes:
                if not isinstance(exclude, str):
                    raise TypeError('The contents of the excludes must all be strings')
                exclude_files.append(exclude)
        fix_excludes = [exclude.replace('.py', '') for exclude in exclude_files]
        excluded_files = tuple(set(files) - set(fix_excludes))
        classes = []
        for file in excluded_files:
            module = importlib.import_module(file)
            for mod_name, clazz in inspect.getmembers(module, self.__context.predicate):
                if hasattr(clazz, _DECORATOR_ATTR) and clazz.load_flg:
                    classes.append(clazz)
                    break
                if self.__context.draw_comparison(file) != mod_name.lower():
                    continue
                classes.append(clazz)
        del self.__context
        has_order_classes = [clazz for clazz in classes if hasattr(clazz, 'load_order') and clazz.load_order]
        if not has_order_classes:
            return tuple(classes)
        no_has_order_classes = [clazz for clazz in classes if not hasattr(clazz, 'load_order') or not clazz.load_order]
        if not no_has_order_classes:
            return tuple(sorted(has_order_classes, key=lambda clazz: clazz.load_order))
        ordered_classes = sorted(has_order_classes, key=lambda clazz: clazz.load_order) + no_has_order_classes
        return tuple(ordered_classes)

    class Context:
        # Don't use enum because it is not supported under Python 3.4 version
        class Type:
            func = 'function'
            clazz = 'class'

        def __init__(self, type):
            self.__type = type
            if type == self.Type.clazz:
                self.__predicate = inspect.isclass
            else:
                self.__predicate = inspect.isfunction

        @property
        def predicate(self):
            return self.__predicate

        def draw_comparison(self, file):
            if self.__type == self.Type.clazz:
                return "".join(file.split("_")).lower()
            else:
                return file.lower()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autoload']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'autoload-module',
    'version': '1.1.1',
    'description': 'Python Autoload Module',
    'long_description': '# autoload_module\nThis library will give you comfortable Python metaprogramming.  \nThe following is a plain example.\n\n- Directory\n```\nproject/\n ├ example.py\n └ validator/\n   ├ validator_a.py\n   ├ validator_b.py\n   └ validator_c.py\n```\n- example.py\n```python\nfrom autoload.module_loader import ModuleLoader\n\ninput = "foo bar baz"\nloader = ModuleLoader()\n\n# Automatically import modules and return class objects\nvalidator_classes = loader.load_classes("validator")\ntry:\n    # initialize and execute method\n    [clazz().validate(input) for clazz in validator_classes]\nexcept:\n    print("input is invalid!!")\n```\n## Install\n```\npip install autoload-module\n```\n## Usage\n### Constructor\n```\nModuleLoader([base_path])\n```\nThe ModuleLoader can be generated with no parameters.\nIn that case, the instance has the absolute path where\nit was initialized.  \n- Directory\n```\n/usr/local/src/project/\n  ├ example.py\n  └ validator/\n    ├ validator_a.py\n    ├ validator_b.py\n    └ validator_c.py\n```\n- example.py\n```python\nfrom autoload.module_loader import ModuleLoader\n\n# The instance has \'/usr/local/src/project/\'\nloader = ModuleLoader()\n\n# load modules in the directory; \'/usr/local/src/project/validator/\'\nvalidator_classes = loader.load_classes("validator")\n```\nIf you want to change the base path, you must generate the ModuleLoader with an absolute path parameter.\n```python\nloader = ModuleLoader(\'/user/local/src/custom\')\n```\n### Methods\n#### load_classes\n```\nload_classes(pkg_name, [excludes])\n```\nThis method read the Python package and return the tuple of class objects.\n- Directory\n```\nproject/\n ├ __init__.py\n ├ config.yaml\n ├ example.py\n ├ validator_a.py\n ├ validator_b.py\n └ validator_c.py\n```\n- validator_a.py\n```python\nclass ValidatorA:\n    def validate(self):\n        print("validateA!!")\n```\n- example.py\n```python\nloader = ModuleLoader()\n\n# Automatically read modules without \'__init__.py\', not py file, and this file.\n# return the tuple of ValidateA, ValidatorB, and ValidatorC class objects\nvalidator_classes = loader.load_classes("project")\n\n# initialize and execute method\n[clazz().validate() for clazz in validator_classes]\n# -> validateA!!\n# -> validateB!!\n# -> validateC!!\n```\nYou can also load only specific modules using `excludes` variable as below.\n```python\n# \'excludes\' is a iterable object like tuple, list.\n# You must specify module names in \'excludes\'.\nvalidator_classes = loader.load_classes("project", ["validator_a", "validator_b"])\n\n[clazz().validate() for clazz in validator_classes]\n# -> validateC!!\n```\nYou can specify `pkg_name` as below.\n```python\nloader.load_classes("validator.py")\nloader.load_classes(".validator")\nloader.load_classes("/validator")\nloader.load_classes("./validator")\n\n# relative path\nloader.load_classes("..packageA.validator")\nloader.load_classes("../packageA/validator")\n```\n\n#### load_functions\n```\nload_functions(pkg_name, [excludes])\n```\nThis method read the Python package and return the tuple of functions.\nThe usage is the same as `load_classes`.\n\n**NOTE**\n- To search class or function, **You must match the name of file and the one of class or function.**\nFor example, if you named the file `test_module.py`, you must named the class `TestModule` or the function `test_module`.\nWhen you want to customize their name, use `@load_config` decorator and write `load=True` manually.\n    - validator_a.py\n    ```python\n    from autoload.decorator import load_config\n  \n    @load_config(load=True)\n    class CustomValidator:\n        def validate(self):\n            print("validateA!!")\n    ```\n- You can also control the order of loaded class objects using `@load_config` decorator.\n    - validator_a.py\n    ```python\n    from autoload.decorator import load_config\n  \n    # sort in ascending order\n    @load_config(order=1)\n    class ValidatorA:\n        def validate(self):\n            print("validateA!!")\n    ```\n#### load_class\n```\nload_class(file_name)\n```\nThis method read the Python file and return the class object.\n- Directory\n```\nproject/\n  ├ example.py\n  └ validator.py\n```\n- validator.py\n```python\nclass Validator:\n    def validate(self):\n        print("validate!!")\n```\n- example.py\n```python\nloader = ModuleLoader()\nclazz = loader.load_class("validator")\nclazz().validate()\n# -> validate!!\n```\nHow to specify `file_name` is the same as that of `load_classes`.\n\n#### load_function\n```\nload_function(file_name)\n```\nThis method read the Python file and return a function object.\nThe usage is the same as `load_class`.\n\n## License\nReleased under the MIT license.',
    'author': 'Hiroki Miyaji',
    'author_email': 'nukoprogramming@gmail.com',
    'maintainer': 'Hiroki Miyaji',
    'maintainer_email': 'nukoprogramming@gmail.com',
    'url': 'https://github.com/hiroki0525/autoload_module',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



def load_config(order=None, load=False):
    def decorator(resource):
        if order:
            resource.load_order = order
        resource.load_flg = load
        return resource
    return decorator